import logging
import reutil

import ccc.github
import checkmarx.util
import cnudie.retrieve
import concourse.steps.scan_sources

__cmd_name__ = 'cx'

logger = logging.getLogger(__name__)


def scan(
    checkmarx_cfg_name: str,
    component: str,
    scan_timeout: int=3600,
    team_id: str=None,
    force: bool=False,
    exclude_regex: [str] = [],
    include_regex: [str] = [],
):
    lookup = cnudie.retrieve.create_default_component_descriptor_lookup()
    component_descriptor = lookup(component)

    concourse.steps.scan_sources.scan_sources(
        checkmarx_cfg_name=checkmarx_cfg_name,
        team_id=team_id,
        component_descriptor=component_descriptor,
        force=force,
        exclude_paths=exclude_regex,
        include_paths=include_regex,
        timeout_seconds=scan_timeout,
    )


def generate_scan_archive(
    repo_url: str, # github.com/gardener/cc-utils
    commit_hash: str | None=None,
    exclude_regex: [str] = [],
    include_regex: [str] = [],
    out_file_path: str='checkmarx_archive',
):
    '''
    create checkmarx scan archive with filters applied
    does not upload to scan API

    commit_hash defaults to repo main branch
    '''

    path_filter_func = reutil.re_filter(
        include_regexes=include_regex,
        exclude_regexes=exclude_regex,
    )

    gh_api = ccc.github.github_api(repo_url=repo_url)
    _, org, repo = repo_url.split('/')
    repo = gh_api.repository(org, repo)

    if not commit_hash:
        commit_hash = repo.default_branch

    logger.info(f'{commit_hash=}')
    logger.info(f'{out_file_path=}')

    with open(out_file_path, 'wb') as f:
        checkmarx.util._download_and_zip_repo(
            clogger=logger,
            repo=repo,
            ref=commit_hash,
            tmp_file=f,
            path_filter_func=path_filter_func,
        )
        f.seek(0)



'''
a collection of hopefully useful tools for compliance-related, half-automated tasks

since some commands will require many (and lengthy) arguments, which, in many cases, will
rarely change, callers may optionally specify a "defaults file" (which is expected to be a YAML
document containing default-values using CLI-arg-names).

Example defaults-file contents:

```
left_name: github.com/example/example
left_version: 1.2.3
right_name: github.com/example/example
right_version: 2.0.0
ocm_repo_urls:
 - eu.gcr.io/example/example-repo
```
'''

import dataclasses
import re

import dacite
import yaml

import gci.componentmodel as cm

import ci.util
import cnudie.util
import cnudie.retrieve
import ctx
import reutil


_cfg = ctx.cfg


@dataclasses.dataclass
class ComponentResourceNames:
    component_name: str
    resource_name: str


@dataclasses.dataclass
class DiffArguments:
    left_name: str
    right_name: str
    left_version: str
    right_version: str
    ocm_repo_urls: list[str]
    exclude_component_names: list[str] = None
    exclude_component_resource_names: list[ComponentResourceNames] = None
    resource_types: list[str] = None
    name_template: str = None
    name_template_expr: str = None
    outfile_prefix: str = 'resource-diff'


def diff(
    left_name: str=None,
    right_name: str=None,
    left_version: str=None,
    right_version: str=None,
    name_template: str=None,
    name_template_expr: str=None,
    resource_types: [str]=None,
    ocm_repo_urls: [str]=None,
    cache_dir: str=_cfg.ctx.cache_dir,
    defaults_file: str=None,
    outfile_prefix: str='resource-diff'
):
    if defaults_file:
        params = ci.util.parse_yaml_file(defaults_file)
    else:
        params = {}

    if left_name:
        params['left_name'] = left_name
    if right_name:
        params['right_name'] = right_name
    if left_version:
        params['left_version'] = left_version
    if right_version:
        params['right_version'] = right_version
    if ocm_repo_urls:
        params['ocm_repo_urls'] = ocm_repo_urls
    if name_template:
        params['name_template'] = name_template
    if name_template_expr:
        params['name_template_expr'] = name_template
    if resource_types:
        params['resource_types'] = resource_types

    try:
        parsed = dacite.from_dict(
            data_class=DiffArguments,
            data=params,
        )
    except:
        print('missing arguments (check either CLI or defaults_file)')
        raise

    if parsed.name_template and parsed.name_template_expr:
        raise ValueError('at most one of name_template_expr, name_template must be specified')

    print('retrieving component-descriptors (might take a few seconds)')

    component_descriptor_lookup = cnudie.retrieve.create_default_component_descriptor_lookup(
        ocm_repository_lookup=cnudie.retrieve.ocm_repository_lookup(
            *parsed.ocm_repo_urls,
        ),
        cache_dir=cache_dir,
    )

    def _components(
        component,
    ):
        if parsed.exclude_component_names:
            component_filter = reutil.re_filter(
                include_regexes=(),
                exclude_regexes=parsed.exclude_component_names,
                value_transformation=lambda comp: comp.name
            )
        else:
            component_filter = None

        return tuple([
            c for c in
            cnudie.retrieve.components(
                component=component,
                component_descriptor_lookup=component_descriptor_lookup,
            )
            if not component_filter or (component_filter and component_filter(c))
        ])

    left_cd = component_descriptor_lookup(cm.ComponentIdentity(
        name=parsed.left_name,
        version=parsed.left_version,
    ))
    left_components = _components(
        component=left_cd,
    )

    right_cd = component_descriptor_lookup(cm.ComponentIdentity(
        name=parsed.right_name,
        version=parsed.right_version,
    ))
    right_components = _components(
        component=right_cd,
    )

    def resource_version_id(component, resource):
        # ignore component-version, honour resource-version
        return component.name, resource.name, resource.version

    def iter_resources_with_ids(components):
        for c in components:
            for r in c.resources:
                if parsed.resource_types:
                    if isinstance(r.type, cm.ResourceType):
                        resource_type = r.type.value
                    else:
                        resource_type = r.type

                    if not resource_type in parsed.resource_types:
                        continue

                if parsed.exclude_component_resource_names:
                    skip = True
                    for component_resource_name in parsed.exclude_component_resource_names:
                        if re.fullmatch(component_resource_name.component_name, c.name) \
                            and \
                            re.fullmatch(component_resource_name.resource_name, r.name):
                            break
                    else:
                        skip = False

                    if skip:
                        continue

                yield c, r, resource_version_id(c, r)

    left_resource_version_ids = {
        cri[2] for cri in iter_resources_with_ids(left_components)
    }
    right_resource_version_ids = {
        cri[2] for cri in iter_resources_with_ids(right_components)
    }

    new_resource_version_ids = [
        (c,r,i) for c,r,i in iter_resources_with_ids(right_components)
        if not i in left_resource_version_ids
    ]
    removed_resource_version_ids = [
        (c,r,i) for c,r,i in iter_resources_with_ids(left_components)
        if not i in right_resource_version_ids
    ]

    def resource_as_dict(component, resource, resource_id):
        if (main_src := cnudie.util.main_source(component, absent_ok=True)):
            src_url = main_src.access.repoUrl
        elif isinstance(resource.access, cm.OciAccess):
            src_url = resource.access.imageReference
        else:
            src_url = '<unknown>'

        if isinstance(resource.access, cm.OciAccess):
            if orig_label := resource.find_label(
                'cloud.gardener.cnudie/migration/original_ref'
            ):
                img_ref = orig_label.value
            else:
                img_ref = resource.access.imageReference

            pull_cmd = {'pull_cmd': f'docker pull {img_ref}'}
        else:
            pull_cmd = {}

        if parsed.name_template:
            name = parsed.name_template.format(
                resource=resource,
                component=component,
            )
        elif parsed.name_template_expr:
            name = eval(parsed.name_template_expr) # nosec B307
        else:
            name = resource.name

        return {
            'name': name,
            'version': resource.version,
            'src_url': src_url,
            **pull_cmd,
        }

    print(f'{left_cd.component.name}:{left_cd.component.version} -> {right_cd.component.version}')
    print(20 * '=')

    print(f'found {len(new_resource_version_ids)=}')
    print('listing new resource-versions')
    print()

    print(yaml.safe_dump((new_resources := [
        resource_as_dict(c,r,i) for c,r,i in new_resource_version_ids
    ])))

    print()
    print()

    print(f'found {len(removed_resource_version_ids)=}')
    print('listing removed resource-versions')
    print()

    print(yaml.safe_dump((removed_resources := [
        resource_as_dict(c,r,i) for c,r,i in removed_resource_version_ids
    ])))

    outfile_new = f'{parsed.outfile_prefix}-added.yaml'
    outfile_removed = f'{parsed.outfile_prefix}-removed.yaml'

    print()

    with open(outfile_new, 'w') as f:
        print(f'writing added resource-versions to {outfile_new=}')
        yaml.safe_dump(
            new_resources,
            f,
        )
        print()

    with open(outfile_removed, 'w') as f:
        print(f'writing removed resource-versions to {outfile_removed=}')
        yaml.safe_dump(
            removed_resources,
            f,
        )



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0

import urllib

from ci.util import (
    ctx,
)
from github.util import (
    GitHubRepositoryHelper,
    find_greatest_github_release_version,
    outdated_draft_releases,

)
import ccc.github

import github3


def list_draft_releases(
    github_cfg_name: str,
    github_repository_owner: str,
    github_repository_name: str,
    only_outdated: bool = False,
):
    '''List all draft releases in a GitHub repository. If the `--only-outdated` flag is set,
    only outdated draft releases are printed. A draft release is considered outdated iff:
        1: its version is smaller than the greatest release version (according to semver) AND
            2a: it is NOT a hotfix draft release AND
            2b: there are no hotfix draft releases with the same major and minor version
            OR
            3a: it is a hotfix draft release AND
            3b: there is a hotfix draft release of greater version (according to semver)
                with the same major and minor version

    Hotfix draft release in this context are draft releases with a semver patch version that is
    not equal to 0.
    '''
    github_cfg = ctx().cfg_factory().github(github_cfg_name)
    github_helper = GitHubRepositoryHelper(
        owner=github_repository_owner,
        name=github_repository_name,
        github_cfg=github_cfg,
    )
    if only_outdated:
        releases = [release for release in github_helper.repository.releases()]
        non_draft_releases = [release for release in releases if not release.draft]
        greatest_release_version = find_greatest_github_release_version(non_draft_releases)
    else:
        releases = github_helper.repository.releases()

    draft_releases = [release for release in releases if release.draft]

    if only_outdated:
        if greatest_release_version is not None:
            draft_releases = outdated_draft_releases(
                draft_releases=draft_releases,
                greatest_release_version=greatest_release_version,
            )
        else:
            draft_releases = []
    for draft_release in draft_releases:
        print(draft_release.name)


def delete_releases(
    github_cfg_name: str,
    github_repository_owner: str,
    github_repository_name: str,
    release_name: [str],
):
    github_cfg = ctx().cfg_factory().github(github_cfg_name)
    github_helper = GitHubRepositoryHelper(
        owner=github_repository_owner,
        name=github_repository_name,
        github_cfg=github_cfg,
    )
    github_helper.delete_releases(release_names=release_name)


def greatest_release_version(
    github_repository_url: str,
    anonymous: bool=False,
    ignore_prereleases: bool=False,
):
    '''Find the release with the greatest name (according to semver) and print its semver-version.

    Note:
    - This will only consider releases whose names are either immediately parseable as semver-
    versions, or prefixed with a single character ('v').
    - The 'v'-prefix (if present) will be not be present in the output.
    - If a release has no name, its tag will be used instead of its name.

    For more details on the ordering of semantic versioning, see 'https://www.semver.org'.
    '''
    parse_result = urllib.parse.urlparse(github_repository_url)

    if not parse_result.netloc:
        raise ValueError(f'Could not determine host for github-url {github_repository_url}')
    host = parse_result.netloc

    try:
        path = parse_result.path.strip('/')
        org, repo = path.split('/')
    except ValueError as e:
        raise ValueError(f"Could not extract org- and repo-name. Error: {e}")

    if anonymous:
        if 'github.com' not in host:
            raise ValueError("Anonymous access is only possible for github.com")
        github_api = github3.GitHub()
        repo_helper = GitHubRepositoryHelper(owner=org, name=repo, github_api=github_api)

    else:
        repo_helper = ccc.github.repo_helper(host=host, org=org, repo=repo)

    print(
        find_greatest_github_release_version(
            releases=repo_helper.repository.releases(),
            warn_for_unparseable_releases=False,
            ignore_prerelease_versions=ignore_prereleases,
        )
    )



import gci.componentmodel as cm

import ccc.clamav
import ccc.delivery
import ccc.oci
import ci.util
import clamav.cnudie
import cnudie.retrieve
import cnudie.iter
import dso.model


__cmd_name__ = 'clamav'


def scan_component(
    component_descriptor_path: str,
    clamav_url: str,
    max_worker: int = 16,
):
    '''
    send component to clamav scanner and write results to stdout.
    '''
    component = cm.ComponentDescriptor.from_dict(
        ci.util.parse_yaml_file(component_descriptor_path)
    ).component

    component_descriptor_lookup = cnudie.retrieve.create_default_component_descriptor_lookup(
        ocm_repository_lookup=cnudie.retrieve.ocm_repository_lookup(
            component.current_repository_ctx(),
        ),
    )

    resource_nodes = cnudie.iter.iter(
        component=component,
        lookup=component_descriptor_lookup,
        node_filter=cnudie.iter.Filter.resources,
    )

    clamav_client = ccc.clamav.client(url=clamav_url)
    oci_client = ccc.oci.oci_client()

    clamav_version_info = clamav_client.clamav_version_info()

    for result in clamav.cnudie.scan_resources(
        resource_nodes=resource_nodes,
        oci_client=oci_client,
        clamav_client=clamav_client,
        clamav_version_info=clamav_version_info,
        max_workers=max_worker,
        virus_db_max_age_days=0,
    ):
        findings_data = clamav.cnudie.resource_scan_result_to_artefact_metadata(
            resource_scan_result=result,
            datasource=dso.model.Datasource.CLAMAV,
            datatype=dso.model.Datatype.MALWARE,
        )
        print(f'{findings_data=}')


def scan_file(
    file_path: str,
    clamav_url: str,
):
    '''
    send file content to clamav scanner and write results to stdout.
    '''

    clamav_client = ccc.clamav.client(url=clamav_url)
    with open(file_path, 'rb') as f:
        contents = f.readlines()

    scan_result = clamav_client.scan(
        data=contents,
    )

    print(scan_result)
